#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library

# import interfaces
from pyams_content.component.paragraph.interfaces import IParagraphContainer, IParagraphContainerTarget, \
    PARAGRAPH_CONTAINER_KEY
from pyams_content.features.checker.interfaces import IContentChecker
from zope.annotation.interfaces import IAnnotations
from zope.location.interfaces import ISublocations
from zope.traversing.interfaces import ITraversable

# import packages
from pyams_content.features.checker import BaseContentChecker
from pyams_utils.adapter import adapter_config, ContextAdapter
from pyams_utils.container import BTreeOrderedContainer
from pyramid.threadlocal import get_current_registry
from zope.interface import implementer
from zope.lifecycleevent import ObjectCreatedEvent
from zope.location import locate

from pyams_content import _


@implementer(IParagraphContainer)
class ParagraphContainer(BTreeOrderedContainer):
    """Paragraphs container"""

    last_id = 1

    def append(self, value):
        key = str(self.last_id)
        self[key] = value
        self.last_id += 1


@adapter_config(context=IParagraphContainerTarget, provides=IParagraphContainer)
def paragraph_container_factory(target):
    """Paragraphs container factory"""
    annotations = IAnnotations(target)
    container = annotations.get(PARAGRAPH_CONTAINER_KEY)
    if container is None:
        container = annotations[PARAGRAPH_CONTAINER_KEY] = ParagraphContainer()
        get_current_registry().notify(ObjectCreatedEvent(container))
        locate(container, target, '++paras++')
    return container


@adapter_config(name='paras', context=IParagraphContainerTarget, provides=ITraversable)
class ParagraphContainerNamespace(ContextAdapter):
    """++paras++ namespace adapter"""

    def traverse(self, name, furtherpath=None):
        return IParagraphContainer(self.context)


@adapter_config(name='paras', context=IParagraphContainerTarget, provides=ISublocations)
class ParagraphContainerSublocations(ContextAdapter):
    """Paragraphs container sublocations"""

    def sublocations(self):
        return IParagraphContainer(self.context).values()


@adapter_config(name='paras', context=IParagraphContainerTarget, provides=IContentChecker)
class ParagraphContainerChecker(BaseContentChecker):
    """Paragraphs container checker"""

    label = _("Paragraphs")
    sep = '\n'
    weight = 10

    def inner_check(self, request):
        output = []
        registry = request.registry
        translate = request.localizer.translate
        visible = False
        for paragraph in IParagraphContainer(self.context).values():
            if not paragraph.visible:
                continue
            visible = True
            for name, checker in sorted(registry.getAdapters((paragraph, ), IContentChecker),
                                        key=lambda x: x[1].weight):
                if name:
                    output.append('<div class="padding-left-20">')
                output.append('- {0} :'.format(translate(checker.label)))
                output.append(checker.get_check_output(request))
                if name:
                    output.append('</div>')
        if not visible:
            output.append('<span class="text-danger">{0}</span>'.format(translate(_("no visible paragraph"))))
        return output
