#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
import json

# import interfaces
from pyams_content.interfaces import MANAGE_CONTENT_PERMISSION
from pyams_content.component.association.interfaces import IAssociationContainer
from pyams_content.component.paragraph.interfaces import IParagraphContainerTarget, IParagraphContainer, IBaseParagraph, \
    IParagraphFactorySettings
from pyams_content.component.paragraph.zmi.interfaces import IParagraphInnerEditor, IParagraphTitleToolbar, \
    IParagraphContainerTable, IParagraphContainerView
from pyams_form.interfaces.form import IFormSecurityContext, IInnerSubForm
from pyams_i18n.interfaces import II18n
from pyams_skin.interfaces import IInnerPage, IPageHeader
from pyams_skin.interfaces.container import ITableElementEditor
from pyams_skin.layer import IPyAMSLayer
from pyams_utils.interfaces import VIEW_SYSTEM_PERMISSION
from pyams_zmi.interfaces.menu import IPropertiesMenu
from pyams_zmi.layer import IAdminLayer
from z3c.table.interfaces import IColumn, IValues
from zope.contentprovider.interfaces import IContentProvider

# import packages
from pyams_content.component.association.zmi import AssociationsContainerView
from pyams_content.shared.common.zmi import WfModifiedContentColumnMixin
from pyams_form.security import ProtectedFormObjectMixin
from pyams_pagelet.pagelet import pagelet_config, Pagelet
from pyams_skin.page import DefaultPageHeaderAdapter
from pyams_skin.table import BaseTable, I18nColumn, TrashColumn, JsActionColumn, SorterColumn, ImageColumn
from pyams_skin.viewlet.menu import MenuItem
from pyams_template.template import template_config
from pyams_utils.adapter import adapter_config, ContextRequestViewAdapter, NullAdapter
from pyams_utils.traversing import get_parent
from pyams_utils.url import absolute_url
from pyams_viewlet.manager import viewletmanager_config, WeightOrderedViewletManager, TemplateBasedViewletManager
from pyramid.exceptions import NotFound
from pyramid.view import view_config
from pyams_viewlet.viewlet import viewlet_config, Viewlet
from pyams_zmi.form import AdminDialogDisplayForm
from pyams_zmi.view import AdminView, ContainerAdminView
from pyramid.decorator import reify
from z3c.form import field
from z3c.table.column import GetAttrColumn
from zope.interface import implementer, Interface

from pyams_content import _


@viewlet_config(name='paragraphs.menu', context=IParagraphContainerTarget, layer=IAdminLayer,
                manager=IPropertiesMenu, permission=VIEW_SYSTEM_PERMISSION, weight=100)
class ParagraphsContainerMenu(MenuItem):
    """Paragraphs container menu"""

    label = _("Paragraphs...")
    icon_class = 'fa-paragraph'
    url = '#paragraphs.html'


#
# Paragraphs container view
#

@implementer(IParagraphContainerTable)
class ParagraphContainerBaseTable(ProtectedFormObjectMixin, BaseTable):
    """Paragraphs container table"""

    id = 'paragraphs_list'
    hide_header = True
    sortOn = None

    @property
    def cssClasses(self):
        classes = ['table', 'table-bordered', 'table-striped', 'table-hover', 'table-tight']
        permission = self.permission
        if (not permission) or self.request.has_permission(permission, self.context):
            classes.append('table-dnd')
        return {'table': ' '.join(classes)}

    @property
    def data_attributes(self):
        attributes = super(ParagraphContainerBaseTable, self).data_attributes
        attributes['table'] = {'id': self.id,
                               'data-ams-plugins': 'pyams_content',
                               'data-ams-plugin-pyams_content-src':
                                   '/--static--/pyams_content/js/pyams_content{MyAMS.devext}.js',
                               'data-ams-location': absolute_url(IParagraphContainer(self.context), self.request),
                               'data-ams-pre-reload': 'PyAMS_content.paragraphs.preReload',
                               'data-ams-post-reload': 'PyAMS_content.paragraphs.postReload',
                               'data-ams-tablednd-drag-handle': 'td.sorter',
                               'data-ams-tablednd-drop-target': 'set-paragraphs-order.json'}
        return attributes


class ParagraphContainerTable(ParagraphContainerBaseTable):
    """Paragraph container base table"""

    @reify
    def values(self):
        return list(super(ParagraphContainerBaseTable, self).values)

    def render(self):
        if not self.values:
            translate = self.request.localizer.translate
            message = translate(_("No currently defined paragraph."))
            permission = self.permission
            if (not permission) or self.request.has_permission(permission, self.context):
                manager = get_parent(self.context, IParagraphFactorySettings)
                if (manager is not None) and not manager.allowed_paragraphs:
                    message = '{0}<br />{1}'.format(message,
                                                    translate(_("Check allowed paragraph types to be able to create "
                                                                "new paragraphs.")))
            return message
        return super(ParagraphContainerBaseTable, self).render()


@adapter_config(context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerBaseTable), provides=IValues)
class ParagraphContainerValues(ContextRequestViewAdapter):
    """Paragraphs container values"""

    @property
    def values(self):
        return IParagraphContainer(self.context).values()


@adapter_config(context=(IBaseParagraph, IPyAMSLayer, ParagraphContainerTable), provides=ITableElementEditor)
class BaseParagraphTableElementEditor(NullAdapter):
    """Base paragraph table element editor"""


@adapter_config(name='sorter', context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerBaseTable),
                provides=IColumn)
class ParagraphContainerSorterColumn(ProtectedFormObjectMixin, SorterColumn):
    """Paragraphs container sorter column"""


@adapter_config(name='show-hide', context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerBaseTable),
                provides=IColumn)
class ParagraphContainerShowHideColumn(ProtectedFormObjectMixin, JsActionColumn):
    """Paragraphs container visibility switcher column"""

    cssClasses = {'th': 'action',
                  'td': 'action switcher'}

    icon_class = 'fa fa-fw fa-eye'
    icon_hint = _("Switch paragraph visibility")

    url = 'PyAMS_content.paragraphs.switchVisibility'

    weight = 5

    def get_icon(self, item):
        if item.visible:
            icon_class = 'fa-eye'
        else:
            icon_class = 'fa-eye-slash text-danger'
        return '<i class="fa fa-fw {icon_class}"></i>'.format(icon_class=icon_class)

    def renderCell(self, item):
        if self.permission and not self.request.has_permission(self.permission, context=item):
            return self.get_icon(item)
        else:
            return super(ParagraphContainerShowHideColumn, self).renderCell(item)


@adapter_config(context=ParagraphContainerShowHideColumn, provides=IFormSecurityContext)
def ShowHideColumnSecurityContextFactory(column):
    """Show/hide column security context factory"""
    return column.table.context


@adapter_config(name='pictogram', context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerBaseTable),
                provides=IColumn)
class ParagraphContainerPictogramColumn(ImageColumn):
    """Paragraph container pictogram column"""

    weight = 6

    def get_icon_class(self, item):
        return item.icon_class

    def get_icon_hint(self, item):
        return self.request.localizer.translate(item.icon_hint)


@viewletmanager_config(name='pyams_paragraph.title_toolbar', context=IBaseParagraph, layer=IPyAMSLayer,
                       view=ParagraphContainerTable, provides=IParagraphTitleToolbar)
@template_config(template='templates/paragraph-title-toolbar.pt', layer=IPyAMSLayer)
@implementer(IParagraphTitleToolbar)
class ParagraphTitleToolbarViewletManager(TemplateBasedViewletManager, WeightOrderedViewletManager):
    """Paragraph title toolbar viewlet manager"""


def getParagraphTitleHints(item, request, table):
    """Get paragraphs column title hints"""
    registry = request.registry
    provider = registry.queryMultiAdapter((item, request, table), IContentProvider,
                                          name='pyams_paragraph.title_toolbar')
    if provider is not None:
        provider.update()
        return provider.render()


@adapter_config(name='name', context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerBaseTable),
                provides=IColumn)
class ParagraphContainerBaseTitleColumn(I18nColumn, WfModifiedContentColumnMixin, GetAttrColumn):
    """Paragraph container base title column"""

    _header = _("Title")

    weight = 50

    def renderCell(self, item):
        return '<span class="title">{0}</span>'.format(super(ParagraphContainerBaseTitleColumn, self).renderCell(item))

    def getValue(self, obj):
        return II18n(obj).query_attribute('title', request=self.request) or ' - - - - - - - -'


@adapter_config(name='name', context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerTable),
                provides=IColumn)
class ParagraphContainerTitleColumn(ParagraphContainerBaseTitleColumn):
    """Paragraph container title column"""

    _header = _("Show/hide all paragraphs")

    def renderHeadCell(self):
        return '<span data-ams-stop-propagation="true"' \
               '      data-ams-click-handler="PyAMS_content.paragraphs.switchAllEditors">' \
               '    <span class="small hint" title="{hint}" data-ams-hint-gravity="e">' \
               '        <i class="fa fa-plus-square-o switch"></i>' \
               '    </span>&nbsp;&nbsp;&nbsp;{title}' \
               '</span>'.format(
                    hint=self.request.localizer.translate(_("Click to open/close all paragraphs editors")),
                    title=super(ParagraphContainerTitleColumn, self).renderHeadCell())

    def renderCell(self, item):
        provider = getParagraphTitleHints(item, self.request, self.table) or ''
        return '<div data-ams-stop-propagation="true" ' \
               '     data-ams-click-handler="PyAMS_content.paragraphs.switchEditor">{provider}' \
               '    <span class="small hint" title="{hint}" data-ams-hint-gravity="e">' \
               '        <i class="fa fa-plus-square-o switch"></i>' \
               '    </span>&nbsp;&nbsp;&nbsp;<span class="title">{title}</span>' \
               '</div>' \
               '<div class="inner-table-form editor margin-x-10 margin-bottom-0"></div>'.format(
                    provider=provider,
                    hint=self.request.localizer.translate(_("Click to open/close paragraph editor")),
                    title=super(ParagraphContainerTitleColumn, self).renderCell(item))


@template_config(template='templates/paragraph-title-icon.pt', layer=IPyAMSLayer)
class ParagraphContainerCounterBase(Viewlet):
    """Paragraph container base counter viewlet"""

    weight = 0
    count = None


@adapter_config(name='trash', context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerBaseTable),
                provides=IColumn)
class ParagraphContainerTrashColumn(ProtectedFormObjectMixin, TrashColumn):
    """Paragraphs container trash column"""


@pagelet_config(name='paragraphs.html', context=IParagraphContainerTarget, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
@implementer(IParagraphContainerView)
class ParagraphContainerView(ContainerAdminView):
    """Paragraphs container view"""

    title = _("Paragraphs list")
    table_class = ParagraphContainerTable


@pagelet_config(name='paragraphs-dialog.html', context=IParagraphContainerTarget, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
class ParagraphContainerDialogView(AdminDialogDisplayForm):
    """Paragraphs container dialog view"""

    title = _("Paragraphs list")
    dialog_class = 'modal-large no-widget-toolbar'
    fieldset_class = 'height-300'

    fields = field.Fields(Interface)


@adapter_config(name='paragraphs', context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerDialogView),
                provides=IInnerSubForm)
@template_config(template='templates/paragraphs.pt', layer=IPyAMSLayer)
@implementer(IParagraphContainerView)
class ParagraphsView(Pagelet):
    """Paragraphs view"""

    weight = 90

    def __init__(self, context, request, view):
        super(ParagraphsView, self).__init__(context, request)
        self.table = ParagraphContainerBaseTable(context, request)
        self.table.view = self

    def update(self):
        super(ParagraphsView, self).update()
        self.table.update()


@adapter_config(context=(IParagraphContainerTarget, IPyAMSLayer, ParagraphContainerView), provides=IPageHeader)
class ParagraphHeaderAdapter(DefaultPageHeaderAdapter):
    """Paragraphs container header adapter"""

    back_url = '#properties.html'
    icon_class = 'fa fa-fw fa-paragraph'


@view_config(name='set-paragraphs-order.json', context=IParagraphContainer, request_type=IPyAMSLayer,
             permission=MANAGE_CONTENT_PERMISSION, renderer='json', xhr=True)
def set_paragraphs_order(request):
    """Update paragraphs order"""
    order = list(map(str, json.loads(request.params.get('names'))))
    request.context.updateOrder(order)
    return {'status': 'success'}


@view_config(name='set-paragraph-visibility.json', context=IParagraphContainer, request_type=IPyAMSLayer,
             permission=MANAGE_CONTENT_PERMISSION, renderer='json', xhr=True)
def set_paragraph_visibility(request):
    """Set paragraph visibility"""
    container = IParagraphContainer(request.context)
    paragraph = container.get(str(request.params.get('object_name')))
    if paragraph is None:
        raise NotFound()
    paragraph = IBaseParagraph(paragraph)
    paragraph.visible = not paragraph.visible
    return {'visible': paragraph.visible}


@view_config(name='get-paragraph-editor.json', context=IParagraphContainer, request_type=IPyAMSLayer,
             permission=VIEW_SYSTEM_PERMISSION, renderer='json', xhr=True)
def get_paragraph_editor(request):
    """Get paragraph editor"""
    container = IParagraphContainer(request.context)
    paragraph = container.get(str(request.params.get('object_name')))
    if paragraph is None:
        raise NotFound()
    registry = request.registry
    editor = registry.queryMultiAdapter((paragraph, request), IParagraphInnerEditor)
    if editor is None:
        editor = registry.queryAdapter(paragraph, IParagraphInnerEditor)
    if editor is not None:
        editor.update()
        return editor.render()


@view_config(name='get-paragraphs-editors.json', context=IParagraphContainer, request_type=IPyAMSLayer,
             permission=VIEW_SYSTEM_PERMISSION, renderer='json', xhr=True)
def get_paragraphs_editors(request):
    """Get all paragraphs inner editors"""
    container = IParagraphContainer(request.context)
    registry = request.registry
    result = {}
    for key, paragraph in container.items():
        editor = registry.queryMultiAdapter((paragraph, request), IParagraphInnerEditor)
        if editor is None:
            editor = registry.queryAdapter(paragraph, IParagraphInnerEditor)
        if editor is not None:
            editor.update()
            result[key] = editor.render()
    return result


#
# Paragraphs associations view
#

@viewlet_config(name='paragraphs-associations.menu', context=IParagraphContainerTarget, layer=IAdminLayer,
                manager=IPropertiesMenu, permission=VIEW_SYSTEM_PERMISSION, weight=101)
class ParagraphsAssociationsMenu(MenuItem):
    """Paragraphs associations container menu"""

    label = _("Associations...")
    icon_class = 'fa-link'
    url = '#paragraphs-associations.html'


@pagelet_config(name='paragraphs-associations.html', context=IParagraphContainerTarget, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
@template_config(template='templates/associations.pt', layer=IPyAMSLayer)
@implementer(IInnerPage)
class ParagraphsAssociationsView(AdminView):
    """Paragraphs associations view"""

    title = _("Paragraphs associations")

    @reify
    def associations(self):
        result = []
        for paragraph in IParagraphContainer(self.context).values():
            associations = IAssociationContainer(paragraph, None)
            if associations is not None:
                view = AssociationsContainerView(paragraph, self.request)
                view.widget_icon_class = 'fa fa-fw {0}'.format(paragraph.icon_class)
                view.title = II18n(paragraph).query_attribute('title', request=self.request) or ' - - - - - - - -'
                result.append(view)
        return result

    def update(self):
        super(ParagraphsAssociationsView, self).update()
        for association in self.associations:
            association.update()
